# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test parameter
"""

import unittest
from joulescope.parameter import Parameter


class TestParameter(unittest.TestCase):

    def test_initialize_keywords_simple(self):
        options = [
            ('there', 0),
            ('world', 1),
            ('you', 2)
        ]
        p = Parameter(name='hello', path='universe', default='world', options=options)
        self.assertEqual('hello', p.name)
        self.assertEqual('universe', p.path)
        self.assertEqual(0, p.str_to_value['there'])
        self.assertEqual(1, p.str_to_value['world'])
        self.assertEqual(2, p.str_to_value['you'])
        with self.assertRaises(KeyError):
            p.str_to_value['__invalid__']
        with self.assertRaises(KeyError):
            p.str_to_value[2]  # does not currently map the value



# Copyright 2022-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .device import Device
from joulescope.parameters_v1 import PARAMETERS_DICT, name_to_value, PARAMETERS_DEFAULTS
import logging


_log = logging.getLogger(__name__)


class DeviceJs110(Device):

    def __init__(self, driver, device_path):
        super().__init__(driver, device_path)
        self._param_map = {
            'sensor_power': None,
            'source': None,
            'i_range': 's/i/range/select',
            'v_range': 's/v/range/select',
            'ovr_to_lsb': None,
            'trigger_source': None,
            'io_voltage': 's/extio/voltage',
            'gpo0': 's/gpo/0/value',
            'gpo1': 's/gpo/1/value',
            'current_lsb': 's/i/lsb_src',
            'voltage_lsb': 's/v/lsb_src',
            'current_ranging': None,  # parse
            'current_ranging_type': 's/i/range/mode',
            'current_ranging_samples_pre': 's/i/range/pre',
            'current_ranging_samples_window': self._on_current_ranging_samples_window,
            'current_ranging_samples_post': 's/i/range/post',
            'buffer_duration': self._on_buffer_duration,
            'reduction_frequency': self._on_reduction_frequency,
            'sampling_frequency': self._on_sampling_frequency,
        }
        self._input_sampling_frequency = 2000000
        self._output_sampling_frequency = 2000000
        self._stream_topics = ['s/i/', 's/v/', 's/p/', 's/i/range/', 's/gpi/0/', 's/gpi/1/']

    def _on_current_ranging_samples_window(self, value):
        if value in ['m', 'n']:
            self.publish('s/i/range/win', value)
        else:
            self.publish('s/i/range/win_sz', int(value))
            self.publish('s/i/range/win', 'manual')

    def _current_ranging_split(self, value):
        if value is None or value in [False, 'off']:
            self.parameter_set('current_ranging_type', 'off')
            return
        parts = value.split('_')
        if len(parts) != 4:
            raise ValueError(f'Invalid current_ranging value {value}')
        for p, v in zip(['type', 'samples_pre', 'samples_window', 'samples_post'], parts):
            self.parameter_set('current_ranging_' + p, v)

    def _on_buffer_duration(self, value):
        self.buffer_duration = value

    def _on_reduction_frequency(self, value):
        scnt = int(2_000_000 / value)
        self.publish('s/stats/scnt', scnt)

    def _on_sampling_frequency(self, value):
        value = int(value)
        self.publish('h/fs', int(value))
        self.output_sampling_frequency = value

    def _config_apply(self, config=None):
        for key, value in PARAMETERS_DEFAULTS.get(config, {}).items():
            self.parameter_set(key, value)

    def parameter_set(self, name, value):
        value_orig = value
        p = PARAMETERS_DICT[name]
        if name == 'current_ranging':
            self._current_ranging_split(value)
            return
        if 'read_only' in p.flags:
            _log.warning('Attempting to set read_only parameter %s', name)
            return
        try:
            value = name_to_value(name, value)
        except KeyError:
            if p.validator is None:
                raise KeyError(f'value {value} not allowed for parameter {name}')
            else:
                value = p.validator(value)
        self._parameters[name] = value
        if not self.is_open:
            self._parameter_set_queue.append((name, value_orig))
            return
        k = self._param_map.get(name)
        if isinstance(k, str):
            _log.info(f'parameter_set({name}, {value}) -> {k}')
            self.publish(k, value)
        elif k is not None:
            _log.info(f'parameter_set({name}, {value})')
            k(value)

    def info(self):
        return {
            'type': 'info',
            'ver': 2,
            'model': self.model,
            'hardware_version': 'H',
            'serial_number': self.serial_number,
        }

    def status(self):
        return {
            'driver': {
                'settings_result': {
                    'value': 0,
                    'units': ''},
                'fpga_frame_counter': {
                    'value': 0,
                    'units': 'frames'},
                'fpga_discard_counter': {
                    'value': 0,
                    'units': 'frames'},
                'sensor_flags': {
                    'value': 0,
                    'format': '0x{:02x}',
                    'units': ''},
                'sensor_i_range': {
                    'value': 0,
                    'format': '0x{:02x}',
                    'units': ''},
                'sensor_source': {
                    'value': 0,
                    'format': '0x{:02x}',
                    'units': ''},
                'return_code': {
                    'value': 0,
                    'format': '{}',
                    'units': '',
                },
            }
        }



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import Structure, c_ubyte, c_ushort


# USB PIPE TYPE
PIPE_TYPE_CONTROL = 0
PIPE_TYPE_ISO = 1
PIPE_TYPE_BULK = 2
PIPE_TYPE_INTERRUPT = 3


def structure_to_repr(x):
    values = ['\n    %s=%r' % (field, getattr(x, field)) for field, _ in x._fields_]
    return "%s(%s)" % (x.__class__.__name__, ','.join(values))


def buffer_to_str(x):
    a = ', '.join(['0x%02x' % z for z in x])
    return '[%s]' % a


class RequestType:  # from USB 2.0 spec chapter 9
    DIRECTION_OUT = (0 << 7)  # Host to device
    DIRECTION_IN = (1 << 7)  # Device to host
    TYPE_STANDARD = (0 << 5)  # Standard request
    TYPE_CLASS = (1 << 5)  # Class-specific request
    TYPE_VENDOR = (2 << 5)  # Vendor-specific request
    TYPE_RESERVED = (3 << 5)  # reserved request
    RECIPIENT_DEVICE = (0 << 0)  # Recipient device
    RECIPIENT_INTERFACE = (1 << 0)  # Recipient interface
    RECIPIENT_ENDPOINT = (2 << 0)  # Recipient endpoint
    RECIPIENT_OTHER = (3 << 0)  # Recipient other

    _DIRECTIONS = ['out', 'in']
    _TYPES = ['standard', 'class', 'vendor']
    _RECIPIENTS = ['device', 'interface', 'endpoint', 'other']

    def __init__(self, value=None, direction=None, type_=None, recipient=None):
        self._directions_s2v = dict([(v, getattr(self, 'DIRECTION_' + v.upper())) for v in RequestType._DIRECTIONS])
        self._directions_v2s = dict([(getattr(self, 'DIRECTION_' + v.upper()), v) for v in RequestType._DIRECTIONS])
        self._types_s2v = dict([(v, getattr(self, 'TYPE_' + v.upper())) for v in RequestType._TYPES])
        self._types_v2s = dict([(getattr(self, 'TYPE_' + v.upper()), v) for v in RequestType._TYPES])
        self._recipients_s2v = dict([(v, getattr(self, 'RECIPIENT_' + v.upper())) for v in RequestType._RECIPIENTS])
        self._recipients_v2s = dict([(getattr(self, 'RECIPIENT_' + v.upper()), v) for v in RequestType._RECIPIENTS])

        if value is not None:
            self.direction = self._directions_v2s[value & 0x80]
            self.type = self._types_v2s[value & 0x60]
            self.recipients = self._recipients_v2s[value & 0x1f]
        else:
            self.direction = direction.lower()
            self.type = type_.lower()
            self.recipient = recipient.lower()
            assert (self.direction in self._directions_s2v)
            assert (self.type in self._types_s2v)
            assert (self.recipient in self._recipients_s2v)

    @property
    def u8(self):
        return self._directions_s2v[self.direction] | self._types_s2v[self.type] | self._recipients_s2v[self.recipient]


class Request:
    """Standard USB requests."""
    GET_STATUS = 0
    CLEAR_FEATURE = 1
    SET_FEATURE = 3
    SET_ADDRESS = 5
    GET_DESCRIPTOR = 6
    SET_DESCRIPTOR = 7
    GET_CONFIGURATION = 8
    SET_CONFIGURATION = 9
    GET_INTERFACE = 10
    SET_INTERFACE = 11
    SYNCH_FRAME = 12


class SetupPacket(Structure):
    _fields_ = [
        ('request_type', c_ubyte),
        ('request', c_ubyte),
        ('value', c_ushort),
        ('index', c_ushort),
        ('length', c_ushort),
    ]

    def __repr__(self):
        return structure_to_repr(self)


class InterfaceDescriptor(Structure):
    _fields_ = [
        ('length', c_ubyte),
        ('descriptor_type', c_ubyte),
        ('interface_number', c_ubyte),
        ('alternate_setting', c_ubyte),
        ('num_endpoints', c_ubyte),
        ('interface_class', c_ubyte),
        ('interface_sub_class', c_ubyte),
        ('interface_protocol', c_ubyte),
        ('interface', c_ubyte)
    ]

    def __repr__(self):
        return structure_to_repr(self)


class ControlTransferResponse:

    def __init__(self, setup_packet, result, data=None):
        self.setup_packet = setup_packet
        self.result = result
        self.data = data

    def __str__(self):
        if self.data is None:
            return 'ControlTransferResponse(result=%r)' % (self.result,)
        else:
            return 'ControlTransferResponse(result=%r, data=%s)' % (
                self.result, buffer_to_str(self.data))

    def __repr__(self):
        if self.data is not None:
            return 'ControlTransferResponse(setup_packet=%r, result=%r)' % (
                self.setup_packet, self.result)
        else:
            return 'ControlTransferResponse(setup_packet=%r, result=%r, data=%s)' % (
                self.setup_packet, self.result, buffer_to_str(self.data))



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the data recorder
"""

import unittest
from joulescope.data_recorder import DataRecorder, DataReader
from joulescope import datafile, data_recorder
from joulescope.v0.stream_buffer import StreamBuffer, usb_packet_factory, usb_packet_factory_signal
from joulescope.v0.calibration import Calibration
import io
import tempfile
import numpy as np
import os
import shutil

PACKET_HEADER_SIZE = 8
SAMPLES_PER_PACKET = (512 - PACKET_HEADER_SIZE) // 4


class TestDataRecorder(unittest.TestCase):

    def setUp(self):
        self._samples_per_reduction_orig = data_recorder.SAMPLES_PER_REDUCTION
        self._reductions_per_tlv_orig = data_recorder.REDUCTIONS_PER_TLV
        self._tempdir = tempfile.mkdtemp()
        self._filename1 = os.path.join(self._tempdir, 'f1.joulescope')

    def tearDown(self):
        data_recorder.SAMPLES_PER_REDUCTION = self._samples_per_reduction_orig
        data_recorder.REDUCTIONS_PER_TLV = self._reductions_per_tlv_orig
        shutil.rmtree(self._tempdir)

    def test_init_with_file_handle(self):
        fh = io.BytesIO()
        d = DataRecorder(fh)
        d.close()
        self.assertGreater(len(fh.getbuffer()), 0)

    def test_init_with_filename(self):
        self.assertFalse(os.path.isfile(self._filename1))
        d = DataRecorder(self._filename1)
        self.assertTrue(os.path.isfile(self._filename1))
        d.close()

    def test_user_data(self):
        expect = ['hello', {'there': 'world'}]
        fh = io.BytesIO()
        d = DataRecorder(fh, user_data=expect)
        d.close()
        fh.seek(0)
        r = DataReader().open(fh)
        self.assertEqual(r.user_data, expect)

    def test_empty_file(self):
        fh = io.BytesIO()
        d = DataRecorder(fh)
        d.close()
        fh.seek(0)
        r = DataReader().open(fh)
        self.assertEqual([0, 0], r.sample_id_range)
        self.assertEqual(1.0, r.sampling_frequency)
        self.assertEqual(1.0, r.input_sampling_frequency)
        self.assertEqual(1.0, r.output_sampling_frequency)
        self.assertEqual(1.0, r.reduction_frequency)
        self.assertEqual(0.0, r.duration)
        self.assertEqual(0, r.voltage_range)
        self.assertEqual(0, len(r.get_reduction(0, 0)))
        self.assertEqual(0, len(r.data_get(0, 0)))
        self.assertEqual(0, r.time_to_sample_id(0.0))
        self.assertEqual(0.0, r.sample_id_to_time(0))
        self.assertIsNone(r.samples_get(0, 0))
        r.close()

    def _create_file(self, packet_index, count=None):
        data_recorder.SAMPLES_PER_REDUCTION = 1000
        data_recorder.REDUCTIONS_PER_TLV = 2
        stream_buffer = StreamBuffer(60.0, [10], 1000.0)
        stream_buffer.suppress_mode = 'off'
        if packet_index > 0:
            data = usb_packet_factory(0, packet_index - 1)
            stream_buffer.insert(data)
            stream_buffer.process()

        fh = io.BytesIO()
        d = data_recorder.DataRecorder(fh)
        d.stream_notify(stream_buffer)
        data = usb_packet_factory(packet_index, count)
        stream_buffer.insert(data)
        stream_buffer.process()
        d.stream_notify(stream_buffer)
        d.close()
        fh.seek(0)

        # dfr = datafile.DataFileReader(fh)
        # dfr.pretty_print()
        # fh.seek(0)
        return fh

    def test_normalize_time_arguments(self):
        fh = self._create_file(0, 2)
        r = DataReader().open(fh)
        self.assertEqual(r.normalize_time_arguments(0, 1), (0, 1))
        self.assertEqual(r.normalize_time_arguments(0.0, 0.01, 'seconds'), (0, 10))
        self.assertEqual(r.normalize_time_arguments(-10, -1), (242, 251))
        self.assertEqual(r.normalize_time_arguments(-0.01, -0.005, 'seconds'), (0, 252))

    def test_user_data_none_when_not_provided(self):
        fh = self._create_file(0, 2)
        r = DataReader().open(fh)
        self.assertEqual(r.user_data, None)

    def test_write_read_direct(self):
        fh = self._create_file(0, 2)
        r = DataReader().open(fh)
        self.assertEqual([0, 252], r.sample_id_range)
        r.raw_processor.suppress_mode = 'off'
        data = r.data_get(0, 252, 1)
        np.testing.assert_allclose(np.arange(0, 252 * 2, 2), data[:, 0]['mean'])

    def test_time_conversion(self):
        fh = self._create_file(0, 2)
        r = DataReader().open(fh)
        self.assertEqual([0, 252], r.sample_id_range)
        self.assertEqual(1000, r.output_sampling_frequency)
        self.assertEqual(1.0, r.reduction_frequency)
        self.assertEqual(0.252, r.duration)
        self.assertEqual(0.0, r.sample_id_to_time(0))
        self.assertEqual(0, r.time_to_sample_id(0))
        self.assertEqual(0.2, r.sample_id_to_time(200))
        self.assertEqual(200, r.time_to_sample_id(0.2))

    def test_write_read_direct_with_offset(self):
        fh = self._create_file(0, 2)
        r = DataReader().open(fh)
        r.raw_processor.suppress_mode = 'off'
        # d = np.right_shift(r.raw(5, 10), 2)
        data = r.data_get(5, 10, 1)
        np.testing.assert_allclose(np.arange(10, 20, 2), data[:, 0]['mean'])

    def test_write_read_direct_with_sample_overscan_before(self):
        fh = self._create_file(1, 3)  # will be samples 120 to 250 (not 126 to 252)
        r = DataReader().open(fh)
        r.raw_processor.suppress_mode = 'off'
        data = r.data_get(0, 140, 1)
        np.testing.assert_allclose(np.arange(252, 532, 2), data[:, 0]['mean'])

    def test_write_read_stats_over_samples(self):
        fh = self._create_file(0, 2)
        r = DataReader().open(fh)
        r.raw_processor.suppress_mode = 'off'
        data = r.data_get(0, 50, 5)
        np.testing.assert_allclose(np.arange(4, 100, 10), data[:, 0]['mean'])

    def test_write_read_stats_over_samples_offset(self):
        fh = self._create_file(0, 2)
        r = DataReader()
        r.raw_processor.suppress_mode = 'off'
        r.open(fh)
        data = r.data_get(7, 50, 10)
        np.testing.assert_allclose(np.arange(23, 90, 20), data[:, 0]['mean'])

    def test_write_read_get_reduction(self):
        fh = self._create_file(0, 32)
        dfr = datafile.DataFileReader(fh)
        dfr.pretty_print()
        fh.seek(0)
        r = DataReader().open(fh)
        data = r.get_reduction(0, 4000)
        np.testing.assert_allclose(np.arange(999, 7000, 2000), data[:, 0]['mean'])

    def test_write_read_get_reduction_offset(self):
        fh = self._create_file(0, 32)
        r = DataReader().open(fh)
        data = r.get_reduction(1000, 4000)
        np.testing.assert_allclose([2999, 4999, 6999], data[:, 0]['mean'])

    def test_write_read_reduction_direct(self):
        fh = self._create_file(0, 32)
        r = DataReader().open(fh)
        data = r.data_get(0, 4000, 1000)
        np.testing.assert_allclose(np.arange(999, 7000, 2000), data[:, 0]['mean'])

    def test_write_read_reduction_indirect(self):
        fh = self._create_file(0, 32)
        r = DataReader().open(fh)
        data = r.data_get(0, 4000, 2000)
        np.testing.assert_allclose([1999, 5999], data[:, 0]['mean'])

    def _create_large_file(self, samples=None):
        """Create a large file.

        :param samples: The total number of samples which will be rounded
            to a full USB packet.
        :return: (BytesIO file, samples)
        """
        sample_rate = 2000000
        packets_per_burst = 128
        bursts = int(np.ceil(samples / (SAMPLES_PER_PACKET * packets_per_burst)))
        stream_buffer = StreamBuffer(1.0, [100], sample_rate)
        samples_total = SAMPLES_PER_PACKET * packets_per_burst * bursts

        fh = io.BytesIO()
        d = DataRecorder(fh)
        d.stream_notify(stream_buffer)
        for burst_index in range(bursts):
            packet_index = burst_index * packets_per_burst
            frames = usb_packet_factory_signal(packet_index, count=packets_per_burst, samples_total=samples_total)
            stream_buffer.insert(frames)
            stream_buffer.process()
            d.stream_notify(stream_buffer)
        d.close()
        fh.seek(0)

        # dfr = datafile.DataFileReader(fh)
        # dfr.pretty_print()
        # fh.seek(0)

        return fh, samples_total

    def test_large_file_from_usb(self):
        sample_count = 2000000 * 2
        fh, sample_count = self._create_large_file(sample_count)
        r = DataReader().open(fh)
        self.assertEqual([0, sample_count], r.sample_id_range)
        reduction = r.get_reduction()
        self.assertEqual(sample_count // 20000, len(reduction))

    def create_sinusoid_data(self, sample_rate, samples):
        x = np.arange(samples, dtype=np.float32)
        x *= (1 / sample_rate)
        data = np.empty(samples * 2, dtype=np.uint16)
        data[0::2] = (2000 * np.sin(2 * np.pi * 1000 * x) + 5000).astype(np.uint16)
        data[1::2] = (2000 * np.cos(2 * np.pi * 42 * x) + 5000).astype(np.uint16)
        np.left_shift(data, 2, out=data)
        data_view = data[1::4]
        np.bitwise_or(data_view, 0x20, out=data_view)
        return data

    def create_sinusoid_file(self, sample_rate, samples):
        cal = Calibration()
        cal.current_offset[:7] = -3000
        cal.current_gain[:7] = [1e-3, 1e-4, 1e-5, 1e-6, 1e-7, 1e-8, 1e-9]
        cal.voltage_offset[:2] = -3000
        cal.voltage_gain[:2] = [1e-3, 1e-4]
        cal.data = cal.save(bytes([0] * 32))

        fh = io.BytesIO()
        d = DataRecorder(fh, calibration=cal)

        stream_buffer = StreamBuffer(1.0, [100], sample_rate)
        stream_buffer.calibration_set(cal.current_offset, cal.current_gain, cal.voltage_offset, cal.voltage_gain)
        d.stream_notify(stream_buffer)
        data = self.create_sinusoid_data(sample_rate, samples)

        chunk_size = (sample_rate // 2) * 2
        for i in range(0, 2 * samples, chunk_size):
            stream_buffer.insert_raw(data[i:(i + chunk_size)])
            stream_buffer.process()
            d.stream_notify(stream_buffer)

        d.close()
        fh.seek(0)
        return fh

    def test_statistics_get(self):
        #fh = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data_recording_01.jls')
        fh = self.create_sinusoid_file(2000000, 400000)
        r = DataReader().open(fh)

        t_start, t_stop = 0.066780, 0.069004
        k_start, k_stop = r.normalize_time_arguments(t_start, t_stop, units='seconds')
        ranges = [
            (0, 1000),         # trivial, direct
            (0, 20000),        # trivial, single reduction
            (100000, 101000),  # offset, direct
            (100000, 120000),  # offset, ex
            (99000, 120000),
            (100000, 121000),
            (99000, 121000),
            (k_start, k_stop),
        ]

        for k_start, k_stop in ranges:
            # print(f'range {k_start}:{k_stop}')
            s1 = r.statistics_get(k_start, k_stop, units='samples')
            k = r.samples_get(k_start, k_stop, units='samples', fields=['current'])
            i_mean = np.mean(k['signals']['current']['value'])
            np.testing.assert_allclose(s1['signals']['current']['µ']['value'], i_mean, rtol=0.0005)
        r.close()

    def test_cache_test(self):
        sample_rate = 2000000
        sample_count = sample_rate * 2
        fh = self.create_sinusoid_file(sample_rate, sample_count)
        r = DataReader().open(fh)
        for step_size in [1111, 2000, 11111, 20000]:
            # print(f'step_size = {step_size}')
            for i in range(0, sample_count - step_size, step_size):
                r.raw_processor.reset()
                s1 = r.statistics_get(i, i + step_size, units='samples')
                k = r.samples_get(i, i + step_size, units='samples', fields=['current'])
                i_mean = np.mean(k['signals']['current']['value'])
                np.testing.assert_allclose(s1['signals']['current']['µ']['value'], i_mean, rtol=0.0005)
        r.close()

    def test_single_sample(self):
        fh = self.create_sinusoid_file(2000000, 400000)
        r = DataReader().open(fh)
        s1 = r.statistics_get(20, 20, units='samples')
        k = r.samples_get(20, 21, units='samples', fields=['current'])
        i_mean = k['signals']['current']['value'][0]
        np.testing.assert_allclose(s1['signals']['current']['µ']['value'], i_mean, rtol=0.0005)

    def test_truncated(self):
        stream_buffer = StreamBuffer(400.0, [10], 1000.0)
        stream_buffer.suppress_mode = 'off'

        fh = io.BytesIO()
        d = DataRecorder(fh)
        d.stream_notify(stream_buffer)
        count = 16
        for idx in range(0, 160, count):
            data = usb_packet_factory(idx, count)
            stream_buffer.insert(data)
            stream_buffer.process()
            d.stream_notify(stream_buffer)
        fh.seek(0)
        #r = datafile.DataFileReader(fh)
        #r.pretty_print()
        r = DataReader().open(fh)

    def test_user_footer_data(self):
        header_data = ['hello', {'there': 'world'}]
        footer_data = ['goodbye', {'for': 'now'}]
        fh = io.BytesIO()
        d = DataRecorder(fh, user_data=header_data)
        d.close(footer_user_data=footer_data)
        fh.seek(0)
        r = DataReader().open(fh)
        self.assertEqual(r.user_data, header_data)
        self.assertEqual(r.footer_user_data, footer_data)


class TestDataRecorderInsert(unittest.TestCase):

    def setUp(self):
        self._tempdir = tempfile.mkdtemp()
        self._filename1 = os.path.join(self._tempdir, 'f1.joulescope')

    def tearDown(self):
        shutil.rmtree(self._tempdir)

    def _create_file_insert(self, packet_index, count_incr, count_total):
        stream_buffer = StreamBuffer(10.0, [10], 1000.0)
        stream_buffer.suppress_mode = 'off'
        if packet_index > 0:
            data = usb_packet_factory(0, packet_index - 1)
            stream_buffer.insert(data)
            stream_buffer.process()

        fh = io.BytesIO()
        d = DataRecorder(fh)
        sample_id = stream_buffer.sample_id_range[1]
        for _ in range(0, count_total, count_incr):
            data = usb_packet_factory(packet_index, count_incr)
            stream_buffer.insert(data)
            stream_buffer.process()
            sample_id_next = stream_buffer.sample_id_range[1]
            d.insert(stream_buffer.samples_get(sample_id, sample_id_next))
            sample_id = sample_id_next
            packet_index += count_incr
        d.close()
        fh.seek(0)

        # dfr = datafile.DataFileReader(fh)
        # dfr.pretty_print()
        # fh.seek(0)
        return fh

    def test_create_single(self):
        fh = self._create_file_insert(0, 2, 2)
        r = DataReader()
        r.raw_processor.suppress_mode = 'off'
        r.open(fh)
        self.assertEqual([0, 252], r.sample_id_range)
        self.assertEqual(1000, r.output_sampling_frequency)
        self.assertEqual(0.05, r.reduction_frequency)
        self.assertEqual(0.252, r.duration)
        self.assertEqual(0.0, r.sample_id_to_time(0))
        self.assertEqual(0, r.time_to_sample_id(0))
        self.assertEqual(0.2, r.sample_id_to_time(200))
        self.assertEqual(200, r.time_to_sample_id(0.2))
        data = r.data_get(6, 10, 1)
        np.testing.assert_allclose(np.arange(12, 20, 2), data[:, 0]['mean'])

    def test_create_big_file(self):
        fh = self._create_file_insert(0, 16, 100000)



# Copyright 2018 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common tools used by :class:`joulescope.usb.api.DeviceDriverApi` implementations."""


import time


class RunUntilDone:

    def __init__(self, timeout, name=''):
        self._timeout = timeout
        self._name = name
        self._value = None
        self._time_start = time.time()

    def __str__(self):
        return 'RunUntilDone(timeout=%r, name=%r)' % (self._timeout, self._name)

    @property
    def value(self):
        return self._value

    @property
    def value_args0(self):
        return self._value[0][0]

    def cbk_fn(self, *args, **kwargs):
        self._value = (args, kwargs)

    def is_done(self):
        if self._value is not None:
            return True
        time_delta = time.time() - self._time_start
        if time_delta > self._timeout:
            raise TimeoutError('RunUntilDone %s: timeout %s > %s' % (self._name, time_delta, self._timeout))
        return False



# Copyright 2017 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Perform functional validation testing on real hardware.
"""

import logging
log = logging.getLogger(__name__)


def control_loopback_wvalue(device, request_id, increment=None):
    """Test writing/reading from a single 16-bit location.

    :param device: The USB device instance.
    :param request_id: The request identifier.
    :param increment: The increment value.  The test runs from 0 to 2**16 - 1.

    The device firmware must implement the request.  On control OUT, the
    firmware must store the 16-bit value.  On a control IN, the firmware
    must return the 16-bit value.
    """
    request_id = int(request_id)
    increment = int(increment) if increment is not None else 1
    for value in range(0, 2 ** 16 - 1, increment):
        if value & 0x3ff == 0:
            log.info('test_control_short %.1f%%', value * 100 / 2 ** 16)
        rv = device.control_transfer_out('device', 'vendor',
                                         request=request_id, value=value, index=0)
        if rv.result != 0:
            raise RuntimeError('test_control_short out %d returned %d' % (value, rv.result))
        rv = device.control_transfer_in('device', 'vendor',
                                        request=request_id, value=value, index=0, length=2)
        if rv.result != 0:
            raise RuntimeError('test_control_short in %d returned %d' % (value, rv.result))
        actual = rv.data[0] + rv.data[1] * 256
        if value != actual:
            raise RuntimeError('test_control_short value mismatch: expected %d, received %d' %
                               (value, actual))


def control_loopback_buffer(device, request_id, max_length, iterations=None):
    """Test writing/reading from a memory buffer 16-bit location.

    :param device: The USB device instance.
    :param request_id: The request identifier.
    :param max_length: The maximum length for each data transfer.  The
        firmware must have a dedicated buffer for at least this many bytes.
    :param iterations: The total number of iterations.

    The device firmware must implement the request.  On control OUT, the
    firmware must store the data to the buffer.  On a control IN, the firmware
    must return the buffer in the data phase.
    """
    request_id = int(request_id)
    iterations = int(iterations) if iterations is not None else 256
    max_length = int(max_length)
    for i in range(iterations):
        if i & 0x3 == 0:
            log.info('test_control_long %.1f%%', i * 100 / iterations)
        for length in range(1, max_length):
            data = bytes(range(length))
            rv = device.control_transfer_out('device', 'vendor',
                                             request=request_id, value=0, index=0, data=data)
            if rv.result != 0:
                raise RuntimeError('test_control_data out %d.%d returned %d' % (i, length, rv.result))
            rv = device.control_transfer_in('device', 'vendor',
                                            request=request_id, value=0, index=0, length=length)
            if rv.result != 0:
                raise RuntimeError('test_control_data in %d.%d returned %d' % (i, length, rv.result))
            if rv.data != data:
                raise RuntimeError('ctest_control_data in %d.%d \n    expected = %s\n    received = %s' %
                                   (i, length, data, rv.data))



# Copyright 2020 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Force all Joulescope bootloaders to run their applications.

This command is normally not necessary.  However, if a user forces
a Joulescope into bootloader mode. then the Joulescope bootloader will
no longer automatically launch the application.  This command forces
all connected Joulescopes running the bootloder to start the application.
"""

from joulescope import scan


def parser_config(p):
    """Force all bootloaders to run the application."""
    return on_cmd


def on_cmd(args):
    for d in scan(name='bootloader'):
        d.open()
        print(f'Found {d}')
        try:
            d.go()
        except Exception:
            print(f'Could not start application on {d}')
    return 0



# Copyright 2022-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .device import Device
from joulescope.parameters_v1 import PARAMETERS_DICT, name_to_value


_I_RANGE_LOOKUP = {
    0x01: '10 A',
    0x02: '10 A',
    0x04: '180 mA',
    0x08: '18 mA',
    0x10: '1.8 mA',
    0x20: '180 µA',
    0x40: '18 µA',
}
_SAMPLING_FREQUENCIES = [
    1, 2, 5, 10, 20, 50, 100, 200, 500,
    1_000, 2_000, 5_000, 10_000, 20_000, 50_000,
    100_000, 200_000, 500_000, 1_000_000,
]


def _signal_bool(value):
    if isinstance(value, str):
        value = value.lower()
    if value in [0, False, '0', 'off', 'disable']:
        return False
    elif value in [1, True, '1', 'on', 'enable']:
        return True
    else:
        raise ValueError('invalid signal level.')


def _version_u32_to_str(v):
    v = int(v)
    major = (v >> 24) & 0xff
    minor = (v >> 16) & 0xff
    patch = v & 0xffff
    return f'{major}.{minor}.{patch}'


class DeviceJs220(Device):

    def __init__(self, driver, device_path):
        super().__init__(driver, device_path)
        self._param_map = {
            'i_range': self._on_i_range,
            'v_range': self._on_v_range,
            'buffer_duration': self._on_buffer_duration,
            'reduction_frequency': self._on_reduction_frequency,
            'sampling_frequency': self._on_sampling_frequency,
            'gpo0': self._on_gpo0,
            'gpo1': self._on_gpo1,
        }
        self._input_sampling_frequency = 2000000
        self._output_sampling_frequency = 1000000
        self._parameters['sampling_frequency'] = self._output_sampling_frequency
        self._stream_topics = ['s/i/', 's/v/', 's/p/', 's/i/range/', 's/gpi/0/', 's/gpi/1/']

    def parameter_set(self, name, value):
        value_orig = value
        p = PARAMETERS_DICT[name]
        if 'read_only' in p.flags:
            self._log.warning('Attempting to set read_only parameter %s', name)
            return
        try:
            if name == 'v_range' and value in ['2V', '2 V', '2', 2]:
                value = '2 V'
            else:
                value = name_to_value(name, value)
        except KeyError:
            if p.validator is None:
                raise KeyError(f'value {value} not allowed for parameter {name}')
            else:
                value = p.validator(value)
        self._parameters[name] = value
        if not self.is_open:
            self._parameter_set_queue.append((name, value_orig))
            return
        k = self._param_map.get(name)
        if k is not None:
            k(value)

    def _on_i_range(self, value):
        if value == 0x80:
            self.publish('s/i/range/mode', 'auto')
        elif value == 0:
            self.publish('s/i/range/mode', 'off')
        else:
            value = _I_RANGE_LOOKUP[value]
            self.publish('s/i/range/select', value)
            self.publish('s/i/range/mode', 'manual')

    def _on_v_range(self, value):
        if value == 'auto':
            self.publish('s/v/range/mode', 'auto')
        else:
            # JS110 current ranges are 5 and 15 V, cannot map to 2 V
            if value in ['2V', '2 V']:
                value = '2 V'
            elif value in ['15V', '15 V', 0]:
                value = '15 V'
            self.publish('s/v/range/select', value)
            self.publish('s/v/range/mode', 'manual')

    def _on_buffer_duration(self, value):
        self.buffer_duration = value

    def _on_reduction_frequency(self, value):
        scnt = int(1_000_000 / value)
        self.publish('s/stats/scnt', scnt)

    def _on_sampling_frequency(self, value):
        value = min(int(value), 1000000)
        if value not in _SAMPLING_FREQUENCIES:
            raise ValueError(f'invalid sampling frequency {value}')
        self._log.info('_on_sampling_frequency %s', value)
        self.publish('h/fs', value)
        self.output_sampling_frequency = value
        self._parameters['sampling_frequency'] = value

    def _on_gpo(self, index, value):
        index = int(index)
        if _signal_bool(value):
            topic = 's/gpo/+/!set'
        else:
            topic = 's/gpo/+/!clr'
        value = 1 << index
        self.publish(topic, value)

    def _on_gpo0(self, value):
        self._on_gpo(0, value)

    def _on_gpo1(self, value):
        self._on_gpo(1, value)

    def _config_apply(self, config=None):
        if config is None or config.lower() == 'auto':
            self.publish('s/i/range/mode', 'auto')
            self.publish('s/v/range/select', '15 V')
            self.publish('s/v/range/mode', 'manual')
        elif config == 'ignore':
            pass  # do nothing
        elif config == 'off':
            for topic in self._stream_topics:
                self.publish(topic + 'ctrl', 0)
        else:
            self._log.warning('Unsupported config %s', config)

    def info(self):
        hardware_version = _version_u32_to_str(self.query('c/hw/version'))
        info = {
            'type': 'info',
            'ver': 2,
            'model': self.model,
            'hardware_version': hardware_version,
            'serial_number': self.serial_number,
            'ctl': {
                'hw': {
                    'rev': _version_u32_to_str(self.query('c/hw/version')),
                    'sn_mcu': self.serial_number,
                    'sn_mfg': self.serial_number,
                    'ver': hardware_version,
                },
                'fw': {
                    'ver': _version_u32_to_str(self.query('c/fw/version')),
                }
            },
            'sensor': {
                'fw': {
                    'ver': _version_u32_to_str(self.query('s/fpga/version')),
                },
                'fpga': {
                    'ver': _version_u32_to_str(self.query('s/fpga/version')),
                },
            },
        }
        return info

    def status(self):
        return {
            'driver': {
                'settings_result': {
                    'value': 0,
                    'units': ''},
                'fpga_frame_counter': {
                    'value': 0,
                    'units': 'frames'},
                'fpga_discard_counter': {
                    'value': 0,
                    'units': 'frames'},
                'sensor_flags': {
                    'value': 0,
                    'format': '0x{:02x}',
                    'units': ''},
                'sensor_i_range': {
                    'value': 0,
                    'format': '0x{:02x}',
                    'units': ''},
                'sensor_source': {
                    'value': 0,
                    'format': '0x{:02x}',
                    'units': ''},
                'return_code': {
                    'value': 0,
                    'format': '{}',
                    'units': '',
                },
            }
        }




# Copyright 2019 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from joulescope.v0.filter_fir import FilterFir
import numpy as np
import time


filter_ma3 = [1.0/3, 1.0/3, 1.0/3]
data_x1 = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
data_y1 = [0, 1.0/3, 1, 2, 3, 4, 5, 6, 7, 8, 9]
data_y2 = [1.0/3, 2, 4, 6, 8]


class TestFilterFir(unittest.TestCase):

    def setUp(self):
        self.y = []

    def cbk(self, y):
        self.y.append(y)

    def test_no_downsample(self):
        f = FilterFir([{'M': 1, 'taps': filter_ma3}])
        f.callback_set(self.cbk)
        for x in data_x1:
            f.process(x)
        np.testing.assert_allclose(data_y1, self.y)

    def test_downsample_by_2(self):
        f = FilterFir([{'M': 2, 'taps': filter_ma3}])
        f.callback_set(self.cbk)
        for x in data_x1:
            f.process(x)
        np.testing.assert_allclose(data_y2, self.y)

    def test_downsample_by_2_width_2(self):
        f = FilterFir([{'M': 2, 'taps': filter_ma3}], width=2)
        f.callback_set(self.cbk)
        x = np.zeros(2, dtype=np.float64)
        for idx in range(len(data_x1) - 1):
            x[0] = data_x1[idx]
            x[1] = data_x1[idx + 1]
            f.process(x)
        expect = np.array(data_y1[1:]).reshape((5, 2))
        np.testing.assert_allclose(expect, self.y)

    def test_performance(self):
        taps = np.ones(504)
        taps /= np.sum(taps)
        width = 1
        f = FilterFir([{'M': 10, 'taps': taps}], width=width)
        x = np.random.randn(2000000, width)
        time_start = time.time()
        f.process(x)
        f.process(x)
        f.process(x)
        time_stop = time.time()
        print(f'duration = {time_stop - time_start}')



# Copyright 2019 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the view
"""

import unittest
from joulescope.view import View
from joulescope.v0.stream_buffer import StreamBuffer, usb_packet_factory
import numpy as np


class TestViewOpenClose(unittest.TestCase):

    def test_open_close(self):
        v = View(stream_buffer=None, calibration=None)
        v.open()
        self.assertEqual((('hello', ), {'one': 1}), v.ping('hello', one=1))
        v.close()


class TestView(unittest.TestCase):

    def setUp(self):
        self.b = StreamBuffer(2.0, [10, 10], sampling_frequency=1000)
        self.v = View(stream_buffer=self.b, calibration=None)
        self.b.insert(usb_packet_factory(0, 2))
        self.b.process()
        self.v.open()

    def tearDown(self):
        self.v.close()

    def test_sample_conversion(self):
        self.assertEqual((('hello', ), {'one': 1}), self.v.ping('hello', one=1))
        self.assertEqual(252, self.v.time_to_sample_id(2.0))
        self.assertEqual(2.0, self.v.sample_id_to_time(252))
        self.assertEqual(152, self.v.time_to_sample_id(1.9))

    def test_statistics_get(self):
        s1 = self.v.statistics_get(-2, -1, units='samples')
        # todo  test



#!/usr/bin/env python3
# Copyright 2018-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Joulescope python setuptools module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
import setuptools
import setuptools.dist
import distutils.cmd
from distutils.errors import DistutilsExecError
import os
import sys

setuptools.dist.Distribution().fetch_build_eggs(['Cython', 'numpy'])

import numpy as np


MYPATH = os.path.dirname(os.path.abspath(__file__))
VERSION_PATH = os.path.join(MYPATH, 'joulescope', 'version.py')


try:
    from Cython.Build import cythonize
    USE_CYTHON = os.path.isfile(os.path.join(MYPATH, 'joulescope', 'v0', 'stream_buffer.pyx'))
except ImportError:
    USE_CYTHON = False


about = {}
with open(VERSION_PATH, 'r', encoding='utf-8') as f:
    exec(f.read(), about)


ext = '.pyx' if USE_CYTHON else '.c'
extensions = [
    setuptools.Extension('joulescope.v0.stream_buffer',
        sources=[
            'joulescope/v0/stream_buffer' + ext,
            'joulescope/v0/native/running_statistics.c',
        ],
        include_dirs=[np.get_include()],
    ),
    setuptools.Extension('joulescope.v0.filter_fir',
        sources=[
            'joulescope/v0/filter_fir' + ext,
            'joulescope/v0/native/filter_fir.c',
        ],
        include_dirs=[np.get_include()],
    ),
    setuptools.Extension('joulescope.v0.pattern_buffer',
        sources=['joulescope/v0/pattern_buffer' + ext],
        include_dirs=[np.get_include()],
    ),
    setuptools.Extension('joulescope.v1.stats',
                         sources=['joulescope/v1/stats' + ext],
                         include_dirs=[np.get_include()],
                         ),
]

if USE_CYTHON:
    from Cython.Build import cythonize
    extensions = cythonize(extensions, compiler_directives={'language_level': '3'})  # , annotate=True)


# Get the long description from the README file
with open(os.path.join(MYPATH, 'README.md'), 'r', encoding='utf-8') as f:
    long_description = f.read()


class CustomBuildDocs(distutils.cmd.Command):
    """Custom command to build docs locally."""

    description = 'Build docs.'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # sphinx-build -b html docs build\docs_html
        # defer import so not all setups require sphinx
        from sphinx.application import Sphinx
        from sphinx.util.console import nocolor, color_terminal
        nocolor()
        source_dir = os.path.join(MYPATH, 'docs')
        target_dir = os.path.join(MYPATH, 'build', 'docs_html')
        doctree_dir = os.path.join(target_dir, '.doctree')
        app = Sphinx(source_dir, source_dir, target_dir, doctree_dir, 'html')
        app.build()
        if app.statuscode:
            raise DistutilsExecError(
                'caused by %s builder.' % app.builder.name)


setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    url=about['__url__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    license=about['__license__'],

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[  # Optional
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Science/Research',

        # Pick your license as you wish
        'License :: OSI Approved :: Apache Software License',

        # Operating systems
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Operating System :: Microsoft :: Windows :: Windows 11',
        'Operating System :: MacOS',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX :: Linux',

        # Supported Python versions
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        
        # Topics
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Embedded Systems',
        'Topic :: Software Development :: Testing',
        'Topic :: System :: Hardware :: Hardware Drivers',
        'Topic :: Utilities',
        'Topic :: System :: Hardware :: Universal Serial Bus (USB)',
    ],

    keywords='joulescope driver',

    packages=setuptools.find_packages(exclude=['native', 'docs', 'test', 'dist', 'build']),
    ext_modules=extensions,
    cmdclass={
        'docs': CustomBuildDocs,
    },
    include_dirs=[],
    include_package_data=True,
    
    # See https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires='~=3.9',

    # See https://packaging.python.org/en/latest/requirements.html
    # https://numpy.org/neps/nep-0029-deprecation_policy.html
    install_requires=[
        'numpy',
        'psutil',
        'pyjls>=0.9.2',
        'pyjoulescope_driver>=1.4.7',
        'python-dateutil>=2.7.3',
        'pymonocypher>=3.1.3.2',
        "pywin32; platform_system=='Windows'",
    ],

    extras_require={
        'dev': ['check-manifest', 'coverage', 'Cython', 'wheel', 'sphinx', 'myst-parser'],
    },   

    entry_points={
        'console_scripts': [
            'joulescope=joulescope.entry_points.runner:run',
        ],
    },
    
    project_urls={
        'Bug Reports': 'https://github.com/jetperch/pyjoulescope/issues',
        'Funding': 'https://www.joulescope.com',
        'Twitter': 'https://twitter.com/joulescope',
        'Source': 'https://github.com/jetperch/pyjoulescope/',
    },
)



# Copyright 2020 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import signal
import time
import logging
from joulescope import scan_require_one
import numpy as np
from queue import Queue, Empty


def parser_config(p):
    """Test Joulescope streaming."""
    return on_cmd


def on_cmd(args):
    device = scan_require_one(name='Joulescope', config='auto')
    quit_ = False
    data_queue = Queue()

    def do_quit(*args, **kwargs):
        nonlocal quit_
        quit_ = 'quit from SIGINT'

    def on_stop(event, message):
        nonlocal quit_
        quit_ = 'quit from device stop'

    def on_update_fn(data):
        data_queue.put(data)

    signal.signal(signal.SIGINT, do_quit)
    view = None
    frame_count = 0
    time_start = time.time()
    try:
        device.open()
        view = device.view_factory()
        view.on_update_fn = on_update_fn
        view.open()
        view.on_x_change('resize', {'pixels': 5801})  # prime number
        device.start(stop_fn=on_stop)
        while not quit_:
            try:
                data = data_queue.get(timeout=0.01)
                view.refresh()
                frame_count += 1
                i = data['signals']['current']['µ']['value']
                idx_mask = np.isfinite(i)
                if np.count_nonzero(idx_mask):
                    # access the data to ensure memory validity
                    stats = {
                        'mean': np.mean(i[idx_mask]),
                        'var': np.mean(data['signals']['current']['σ2']['value']),  # not really
                        'p2p': data['signals']['current']['max']['value'] - data['signals']['current']['min']['value']
                    }
            except Empty:
                time.sleep(0.001)
        device.stop()
    except Exception as ex:
        logging.getLogger().exception('while streaming data')
        print('Data streaming failed')
        return 1
    finally:
        if view is not None:
            view.close()
        device.close()
    time_stop = time.time()
    duration = time_stop - time_start
    fps = frame_count / duration
    print(f'done streaming data: {frame_count} frames in {duration:.3f} seconds = {fps:.1f} fps')
    return 0



# Copyright 2022-2023 Jetperch LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The pyjoulescope_driver wrapper to implement the v0 API."""


from pyjoulescope_driver import Driver
from .device import Device
from .js220 import DeviceJs220
from .js110 import DeviceJs110
import atexit
import logging
from typing import List


_log = logging.getLogger(__name__)


class DriverWrapper:
    """Singleton to wrap pyjoulescope_driver.Driver"""
    __singleton__ = None

    def __new__(cls, *args, **kwds):
        s = cls.__dict__.get("__singleton__")
        if s is not None:
            return s
        s = object.__new__(cls)
        cls.__singleton__ = s
        s._initialize(*args, **kwds)
        return s

    def _initialize(self):
        self.driver = Driver()
        self.driver.log_level = 'INFO'
        atexit.register(self._finalize)
        self.devices = {}
        self.driver.subscribe('@/!add', 'pub', self._on_device_add)
        self.driver.subscribe('@/!remove', 'pub', self._on_device_remove)
        for d in self.driver.device_paths():
            self._on_device_add('@/!add', d)

    def _finalize(self):
        while len(self.devices):
            _, device = self.devices.popitem()
            device.close()
        d, self.driver = self.driver, None
        d.finalize()

    def _on_device_add(self, topic, value):
        if value in self.devices:
            return
        if 'js220' in value:
            cls = DeviceJs220
        elif 'js110' in value:
            cls = DeviceJs110
        else:
            _log.info('Unsupported device: %s', value)
            return
        self.devices[value] = cls(self.driver, value)

    def _on_device_remove(self, topic, value):
        d = self.devices.pop(value, None)
        if d is not None:
            d.close()

    def scan(self, name: str = None, config=None):
        if name is None or name.lower() == 'joulescope':
            devices = self.devices.values()
            devices = [d for d in devices if d.device_path[2] != '&']
            devices = sorted(devices, key=lambda x: str(x))
            for d in devices:
                d.config = config
        elif name.lower() == 'bootloader':
            devices = self.devices.values()
            devices = [d for d in devices if d.device_path[2] == '&']
            devices = sorted(devices, key=lambda x: str(x))
            for d in devices:
                d.config = config
        return devices


def scan(name: str = None, config=None) -> List[Device]:
    """Scan for connected devices.

    :param name: The case-insensitive device name to scan.
        None (default) is equivalent to 'Joulescope'.
    :param config: The configuration for the :class:`Device`.
    :return: The list of :class:`Device` instances.  A new instance is created
        for each detected device.  Use :func:`scan_for_changes` to preserved
        existing instances.
    :raises: None - guaranteed not to raise an exception
    """
    d = DriverWrapper()
    return d.scan(name, config)


def scan_require_one(name: str = None, config=None) -> Device:
    """Scan for one and only one device.

    :param name: The case-insensitive device name to scan.
        None (default) is equivalent to 'Joulescope'.
    :param config: The configuration for the :class:`Device`.
    :return: The :class:`Device` found.
    :raise RuntimeError: If no devices or more than one device was found.
    """
    devices = scan(name, config=config)
    if not len(devices):
        raise RuntimeError("no devices found")
    if len(devices) > 1:
        raise RuntimeError("multiple devices found")
    return devices[0]


def scan_for_changes(name: str = None, devices=None, config=None):
    """Scan for device changes.

    :param name: The case-insensitive device name to scan.
        None (default) is equivalent to 'Joulescope'.
    :param devices: The list of existing :class:`Device` instances returned
        by a previous scan.  Pass None or [] if no scan has yet been performed.
    :param config: The configuration for the :class:`Device` which is one of
        ['auto', 'ignore', 'off'].  None is equivalent to 'auto'.
    :return: The tuple of lists (devices_now, devices_added, devices_removed).
        "devices_now" is the list of all currently connected devices.  If the
        device was in "devices", then return the :class:`Device` instance from
        "devices".
        "devices_added" is the list of connected devices not in "devices".
        "devices_removed" is the list of devices in "devices" but not "devices_now".
    """
    devices_prev = [] if devices is None else devices
    devices_next = scan(name, config=config)
    devices_added = []
    devices_removed = []
    devices_now = []

    for d in devices_next:
        matches = [x for x in devices_prev if str(x) == str(d)]
        if len(matches):
            devices_now.append(matches[0])
        else:
            devices_added.append(d)
            devices_now.append(d)

    for d in devices_prev:
        matches = [x for x in devices_next if str(x) == str(d)]
        if not len(matches):
            devices_removed.append(d)

    _log.info('scan_for_changes %d devices: %d added, %d removed',
              len(devices_now), len(devices_added), len(devices_removed))
    return devices_now, devices_added, devices_removed


class DeviceNotify:

    def __init__(self, cbk):
        """Start device insertion/removal notification.

        :param cbk: The function called on device insertion or removal.  The
            arguments are (inserted, info).  "inserted" is True on insertion
            and False on removal.  "info" contains platform-specific details
            about the device.  In general, the application should rescan for
            relevant devices.
        """
        self._on_add_fn = self._on_add
        self._on_remove_fn = self._on_remove
        self._cbk = cbk
        self._driver = None
        self._cbk(True, None)
        self.open()

    def _on_add(self, topic, value):
        self._cbk(True, value)

    def _on_remove(self, topic, value):
        self._cbk(False, value)

    def open(self):
        self.close()
        self._driver = DriverWrapper()
        d = self._driver.driver
        d.subscribe('@/!add', 'pub', self._on_add_fn)
        d.subscribe('@/!remove', 'pub', self._on_remove_fn)

    def close(self):
        if self._driver is not None:
            d = self._driver.driver
            self._driver = None
            d.unsubscribe('@/!add', self._on_add_fn)
            d.unsubscribe('@/!remove', self._on_remove_fn)
